MODULE f90_utilities
!**********************************************************************
!   FORTRAN 90 module for defining necessary utilities for various
!   mathematical computations.
!
!   Created: 10/28/2002 JLH
!
!   Notes:  Some subroutines require BLAS subroutines xGEMV, xGEMM,
!           or BLAS function xdot.  All are available for free download
!           from http://www.netlib.org/blas/ .
!           Some subroutines require LAPACK subroutines xGETRF and
!           xGETRI.  Both are available for free download from
!           http://www.netlib.org/lapack/ .
!           In notation above, the lower-case "x" preceding the
!           subroutine or function name is replaced by an "S" for
!           single precision calls or a "D" for double precision
!           calls.
!
!   Contents:  
!         Routines that move data:
!                       swap
!         Argument checking and error handling: 
!			assert_eq, error
!         Routines for matrix opertaions:
!                       mat_mult, vech, cholesky_fac
!
!**********************************************************************
  USE define_types
  IMPLICIT NONE
!**********************************************************************
!   Routines for moving data:
!**********************************************************************
  INTERFACE swap
     MODULE PROCEDURE swap_i, swap_r, swap_rv, swap_d, swap_dv,       &
          swap_c, swap_cv, swap_cm, swap_z, swap_zv, swap_zm,         &
          masked_swap_rs, masked_swap_rv, masked_swap_rm,             &
          masked_swap_rd, masked_swap_dv, masked_swap_dm               
  END INTERFACE swap
!***********************************************************************
!   Routines for multiplying matrices:
!***********************************************************************
  INTERFACE mat_mult
     MODULE PROCEDURE mat_mult_s, mat_mult_d, vv_mult_s, vv_mult_d,   &
          vm_mult_s, vm_mult_d, mv_mult_s, mv_mult_d
  END INTERFACE mat_mult
!***********************************************************************
!  Routines for computing the half-stacking vector:
!***********************************************************************
  INTERFACE vech
     MODULE PROCEDURE vech_r, vech_d
  END INTERFACE vech
!***********************************************************************
!  Routines for computing the Cholesky decomposition:
!***********************************************************************
  INTERFACE cholesky_fac
     MODULE PROCEDURE cholesky_fac_r, cholesky_fac_d
  END INTERFACE cholesky_fac
!**********************************************************************
!  Routines for argument checking and error handling (error is not
!  currently overloaded.
!**********************************************************************
  INTERFACE assert_eq
     MODULE PROCEDURE assert_eq2, assert_eq3, assert_eq4, assert_eqn
  END INTERFACE assert_eq
!***********************************************************************
!  Routines for sorting arrays and carrying indices.
!***********************************************************************
  INTERFACE index_sort
     MODULE PROCEDURE i_sort, r_sort, d_sort
  END INTERFACE index_sort
CONTAINS
!**********************************************************************
!   Swap the contents of a and b.
!**********************************************************************
  SUBROUTINE swap_i(a,b)
    USE define_types
    IMPLICIT NONE
    INTEGER(i4b), INTENT(inout) :: a, b
    INTEGER(i4b) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_i
!
  SUBROUTINE swap_r(a,b)
    USE define_types
    IMPLICIT NONE
    REAL(sp), INTENT(inout) :: a, b
    REAL(sp) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_r
!
  SUBROUTINE swap_rv(a,b)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:), INTENT(inout) :: a, b
    REAL(sp), DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_rv
!
  SUBROUTINE swap_d(a,b)
    USE define_types
    IMPLICIT NONE
    REAL(dp), INTENT(inout) :: a, b
    REAL(dp) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_d
!
  SUBROUTINE swap_dv(a,b)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(inout) :: a, b
    REAL(dp), DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_dv
!
  SUBROUTINE swap_c(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(spc), INTENT(inout) :: a, b
    COMPLEX(spc) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_c
!
  SUBROUTINE swap_cv(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(spc), DIMENSION(:), INTENT(inout) :: a, b
    COMPLEX(spc), DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_cv
!
  SUBROUTINE swap_cm(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(spc), DIMENSION(:,:), INTENT(inout) :: a, b
    COMPLEX(spc), DIMENSION(SIZE(a,1), SIZE(a,2)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_cm
!
  SUBROUTINE swap_z(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(dpc), INTENT(inout) :: a, b
    COMPLEX(dpc) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_z
!
  SUBROUTINE swap_zv(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(dpc), DIMENSION(:), INTENT(inout) :: a, b
    COMPLEX(dpc), DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_zv
!
  SUBROUTINE swap_zm(a,b)
    USE define_types
    IMPLICIT NONE
    COMPLEX(dpc), DIMENSION(:,:), INTENT(inout) :: a, b
    COMPLEX(dpc), DIMENSION(SIZE(a,1), SIZE(a,2)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_zm
!
  SUBROUTINE masked_swap_rs(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(sp), INTENT(inout) :: a, b
    LOGICAL(lgt), INTENT(in) :: mask
    REAL(sp) :: swp
    IF(mask) THEN
       swp = a
       a = b
       b = swp
    END IF
    RETURN
  END SUBROUTINE masked_swap_rs
!
  SUBROUTINE masked_swap_rv(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:), INTENT(inout) :: a, b
    LOGICAL(lgt), DIMENSION(:), INTENT(in) :: mask
    REAL(sp), DIMENSION(SIZE(a)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_rv
!
  SUBROUTINE masked_swap_rm(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:,:), INTENT(inout) :: a, b
    LOGICAL(lgt), DIMENSION(:,:), INTENT(in) :: mask
    REAL(sp), DIMENSION(SIZE(a,1),SIZE(a,2)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_rm
!
  SUBROUTINE masked_swap_rd(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(dp), INTENT(inout) :: a, b
    LOGICAL(lgt), INTENT(in) :: mask
    REAL(dp) :: swp
    IF(mask) THEN
       swp = a
       a = b
       b = swp
    END IF
    RETURN
  END SUBROUTINE masked_swap_rd
!
  SUBROUTINE masked_swap_dv(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(inout) :: a, b
    LOGICAL(lgt), DIMENSION(:), INTENT(in) :: mask
    REAL(dp), DIMENSION(SIZE(a)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_dv
!
  SUBROUTINE masked_swap_dm(a,b,mask)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:,:), INTENT(inout) :: a, b
    LOGICAL(lgt), DIMENSION(:,:), INTENT(in) :: mask
    REAL(dp), DIMENSION(SIZE(a,1),SIZE(a,2)) ::swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_dm
!***********************************************************************
!   Matrix multiplication routines:
!***********************************************************************
  SUBROUTINE mat_mult_s(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:,:), INTENT(in) :: a, b
    REAL(sp), DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER(i4b) :: n, m, k, info
!
    EXTERNAL sgemm
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    k = SIZE(b,2)
!
    info = assert_eq(SIZE(b,1), m,                                     &
         "Matrix dimensions not correct for multiplication")
!
    CALL sgemm("N","N",n,k,m,1.0_sp,a,n,b,m,1.0_sp,c,n)
!
    RETURN
  END SUBROUTINE mat_mult_s
!
  SUBROUTINE mat_mult_d(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:,:), INTENT(in) :: a, b
    REAL(dp), DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER(i4b) :: n, m, k, info
!
    EXTERNAL dgemm
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    k = SIZE(b,2)
!
    info = assert_eq(SIZE(b,1),m,                                     &
         "Matrix dimensions not correct for multiplication.")
!
    CALL dgemm("N","N",n,k,m,1.0_dp,a,n,b,m,1.0_dp,c,n)
!
    RETURN
  END SUBROUTINE mat_mult_d
!
  SUBROUTINE vv_mult_s(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:), INTENT(in) :: a, b
    REAL(sp), INTENT(out) :: c
    REAL(sp) :: sdot
!
    EXTERNAL sdot
!
    c = sdot(SIZE(a),a,1,b,1)
!
    RETURN
  END SUBROUTINE vv_mult_s
!
  SUBROUTINE vv_mult_d(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(in) :: a, b
    REAL(dp), INTENT(out) :: c
    REAL(dp) :: ddot
!
    EXTERNAL ddot
!
    c = ddot(SIZE(a),a,1,b,1)
!
    RETURN
  END SUBROUTINE vv_mult_d
!
  SUBROUTINE vm_mult_s(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:), INTENT(in) :: a
    REAL(sp), DIMENSION(:,:), INTENT(in) :: b
    REAL(sp), DIMENSION(:,:), INTENT(out) :: c
!
    REAL(sp), DIMENSION(SIZE(a),1) :: work
!
    INTEGER(i4b) :: info, n, k
!
    EXTERNAL sgemm
!
    n = SIZE(a)
    k = SIZE(b,2)
    info = assert_eq(SIZE(b,1),1,"The matrix must be 1 x k.")
!
    work = RESHAPE(a,(/n, 1/))
!
    CALL sgemm("N","N",n,k,1,1.0_sp,work,n,b,1,1.0_sp,c,n)
!
    RETURN
  END SUBROUTINE vm_mult_s
!
  SUBROUTINE vm_mult_d(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:), INTENT(in) :: a
    REAL(dp), DIMENSION(:,:), INTENT(in) :: b
    REAL(dp), DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER(i4b) :: info, n, k
    REAL(dp), DIMENSION(SIZE(a),1) :: work
!
    EXTERNAL dgemm
!
    n = SIZE(a)
    k = SIZE(b,2)
    info = assert_eq(SIZE(b,1),1,"The matrix must be 1 x k.")
!
    work = RESHAPE(a,(/n, 1/))
!
    CALL dgemm("N","N",n,k,1,1.0_dp,work,n,b,1,1.0_dp,c,n)
!
    RETURN
  END SUBROUTINE vm_mult_d
!
  SUBROUTINE mv_mult_s(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:,:), INTENT(in) :: a
    REAL(sp), DIMENSION(:), INTENT(in) :: b
    REAL(sp), DIMENSION(:), INTENT(out) :: c
!
    INTEGER(i4b) :: info, n, m
!
    EXTERNAL sgemv
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    info = assert_eq(m,SIZE(b),"The matrix must be 1 x m.")
!
    CALL sgemv("N",n,m,1.0_sp,a,n,b,1,1.0_sp,c,1)

    RETURN
  END SUBROUTINE mv_mult_s
!
  SUBROUTINE mv_mult_d(a,b,c)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:,:), INTENT(in) :: a
    REAL(dp), DIMENSION(:), INTENT(in) :: b
    REAL(dp), DIMENSION(:), INTENT(out) :: c
!
    INTEGER(i4b) :: info, n, m
!
    EXTERNAL dgemv
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    info = assert_eq(m,SIZE(b),"The matrix must be 1 x m.")
!
    CALL dgemv("N",n,m,1.0_dp,a,n,b,1,1.0_dp,c,1)
!
    RETURN
  END SUBROUTINE mv_mult_d
!***********************************************************************
!   Half-stacking vector subroutines:
!***********************************************************************
!***********************************************************************
!   FORTRAN subroutine to calculate the stacking or half-stacking
!   vector, v, of the product of two vectors, a and b.
!
!   Input: a = a double precision matrix of dimension n x m.
!          b = a double precision matrix of dimension m x k.
!          option = an integer indicating whether to take the half-
!                   stacking vector (iopt = 0) or the stacking vector
!                   (iopt = 1).
!          n = an integer containing the number of columns of a.
!          k = an integer containing the number of rows of b.
!
!   Output: v = a double precision vector containing the half-stacking
!               or stacking vector of a and b.  If iopt = 0, the 
!               length of y is (nk)[(nk)+1]/2.  If iopt = 1, the 
!               length of z is nk.
!***********************************************************************
  SUBROUTINE vech_r(a,b,n,ndim,m,mdim,k,option,v)
    USE define_types
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: n, ndim, m, mdim, k, option
    REAL(sp), DIMENSION(ndim), INTENT(in) :: a
    REAL(sp), DIMENSION(1,mdim), INTENT(in) :: b
    REAL(sp), DIMENSION(k**2), INTENT(inout) :: v
!  
    INTEGER(i4b) :: lower_limit, vindex, i, j
!
    v = 0.0_sp
!
    lower_limit = 1
    vindex = 0
!
    DO i = 1, k
       IF(option == 0) lower_limit = i
       DO j = lower_limit, n
          vindex = vindex + 1
          v(vindex) = a(j)*b(1,i)
       END DO
    END DO
    RETURN
  END SUBROUTINE vech_r
!
  SUBROUTINE vech_d(a,b,n,ndim,m,mdim,k,option,v)
    USE define_types
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: n, ndim, m, mdim, k, option
    REAL(dp), DIMENSION(ndim), INTENT(in) :: a
    REAL(dp), DIMENSION(1,mdim), INTENT(in) :: b
    REAL(dp), DIMENSION(k**2), INTENT(inout) :: v
!
    INTEGER(i4b) :: lower_limit, vindex, i, j
!
    v = 0.0_dp
    
    lower_limit = 1
    vindex = 0
!
    DO i = 1, k
       IF(option == 0) lower_limit = i
       DO j = lower_limit, n
          vindex = vindex + 1
          v(vindex) = a(j)*b(1,i)
       END DO
    END DO
    RETURN
  END SUBROUTINE vech_d
!***********************************************************************
!   Cholesky decomposition subroutines:
!***********************************************************************
!**********************************************************************
!   FORTRAN 90 subroutine for computing the Cholesky factorization of
!   an n x n positive definite matrix a.  The Cholesky factor is 
!   returned in the upper triangular matrix u.
!
!   Created: 10/28/2003 JLH
!**********************************************************************
  SUBROUTINE cholesky_fac_r(x,u)
    USE define_types
    IMPLICIT NONE
    REAL(sp), DIMENSION(:,:), INTENT(in) :: x
    REAL(sp), DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
!
    INTEGER(i4b) :: info, i, j
!
    EXTERNAL spotrf
!
    DO i = 2, SIZE(x,1)
    DO j = 1, i-1
       u(i,j) = 0.0_sp
    END DO
    END DO
!
    CALL spotrf("U",SIZE(x,1),u,SIZE(x,1),info)
!
    RETURN
  END SUBROUTINE cholesky_fac_r
!
  SUBROUTINE cholesky_fac_d(x,u)
    USE define_types
    IMPLICIT NONE
    REAL(dp), DIMENSION(:,:), INTENT(in) :: x
    REAL(dp), DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
!
    INTEGER(i4b) :: info, i, j
!
    EXTERNAL dpotrf
!
    DO i = 2, SIZE(x,1)
    DO j = 1, i-1
       u(i,j) = 0.0_dp
    END DO
    END DO
!
    CALL dpotrf("U",SIZE(x,1),u,SIZE(x,1),info)
!
    RETURN
  END SUBROUTINE cholesky_fac_d
!***********************************************************************
!   Error handling routines:
!**********************************************************************
!   Report and die if integers not all equal (used for size checking).
!**********************************************************************
  FUNCTION assert_eq2(n1,n2,string)
    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER(i4b), INTENT(in) :: n1, n2
    INTEGER(i4b) :: assert_eq2
!
    IF(n1 == n2) THEN
       assert_eq2 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:", string
       STOP " Program terminated by assert_eq2."
    END IF
    RETURN
  END FUNCTION assert_eq2
!
  FUNCTION assert_eq3(n1,n2,n3,string)
    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER(i4b), INTENT(in) :: n1, n2, n3
    INTEGER(i4b) :: assert_eq3
    IF(n1 == n2 .AND. n2 == n3) THEN
       assert_eq3 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eq3."
    END IF
    RETURN
  END FUNCTION assert_eq3
!
  FUNCTION assert_eq4(n1,n2,n3,n4,string)
    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER(i4b), INTENT(in) :: n1, n2, n3, n4
    INTEGER(i4b) :: assert_eq4
    IF(n1 == n2 .AND. n2 == n3 .AND. n3 == n4) THEN
       assert_eq4 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eq4."
    END IF
    RETURN
  END FUNCTION assert_eq4
!
  FUNCTION assert_eqn(nn,string)
    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER(i4b), DIMENSION(:), INTENT(in) :: nn
    INTEGER(i4b) :: assert_eqn
    IF(ALL(nn(2:) == nn(1))) THEN
       assert_eqn = nn(1)
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eqn."
    END IF
    RETURN
  END FUNCTION assert_eqn
!
  SUBROUTINE error(string)
!***********************************************************************
!   Report a message, then die.
!***********************************************************************
    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    WRITE(*,*) "Error: ", string
    STOP " Program terminated by error."
    RETURN
  END SUBROUTINE error
!
  SUBROUTINE i_sort(n,x_in,indices,x_out)
    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER(i4b), INTENT(in) :: n
    INTEGER(i4b), DIMENSION(n), INTENT(inout) :: indices
    INTEGER(i4b), DIMENSION(n), INTENT(in) :: x_in
    INTEGER(i4b), DIMENSION(n), INTENT(out) :: x_out
    INTEGER(i4b) :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE i_sort
!
  SUBROUTINE r_sort(n,x_in,indices,x_out)
    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER(i4b), INTENT(in) :: n
    INTEGER(i4b), DIMENSION(n), INTENT(inout) :: indices
    REAL(sp), DIMENSION(n), INTENT(in) :: x_in
    REAL(sp), DIMENSION(n), INTENT(out) :: x_out
    INTEGER(i4b) :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE r_sort
!
  SUBROUTINE d_sort(n,x_in,indices,x_out)
    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER(i4b), INTENT(in) :: n
    INTEGER(i4b), DIMENSION(n), INTENT(inout) :: indices
    REAL(dp), DIMENSION(n), INTENT(in) :: x_in
    REAL(dp), DIMENSION(n), INTENT(out) :: x_out
    INTEGER(i4b) :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE d_sort
!
END MODULE f90_utilities
