MODULE nlts_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used nonlinear time series
!   statistical computations.
!
!   Created: 03/12/03 JLH
!
!   Contents:
!        Subroutines for generating nonlinear vector realizations:
!            get_mvbilinear_params, gen_mvbilinear, get_mvTAR_params,
!            gen_mvTAR, get_mvSTAR_params, gen_mvSTAR,
!            get_mvEXPAR_params, gen_mvEXPAR
!***********************************************************************
  USE define_types
  USE stat_utilities
!
  IMPLICIT NONE
CONTAINS
!***********************************************************************
!  Generate vector bilinear:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvbilinear_params(max_k,max_p,max_q,file_label,k,p,q,  &
       cp_option,AR_coefs,MA_coefs,cp_dim,cp_coefs)
!***********************************************************************
!  FORTRAN90 subroutine to get parameter values for generating 
!  multivariate bilinear time series.
!***********************************************************************
    USE define_types
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_k, max_p, max_q, file_label, k
    INTEGER(i4b), INTENT(out) :: p, q, cp_option, cp_dim
    REAL(dp), INTENT(out), DIMENSION(max_k, (max_k*max_p)**2) :: cp_coefs
    REAL(dp), INTENT(out), DIMENSION(max_p, max_k, k) :: AR_coefs
    REAL(dp), INTENT(out), DIMENSION(max_q, max_k, k) :: MA_coefs
!
    INTEGER(i4b) :: lag, dim1, dim2
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,500) p, q
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR coefficients, row by row."
       WRITE(file_label,510) 
       DO lag = 1, p
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "Enter the MA coefficients, row by row."
       WRITE(file_label,530) 
       DO lag = 1, q
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    WRITE(*,*) "What type of cross-product terms should be included?"
    WRITE(*,*) " 0 => no cross-product terms,"
    WRITE(*,*) " 1 => time series realizations with errors,"
    WRITE(*,*) " 2 => squared terms and interactions of time series, or"
    WRITE(*,*) " 3 => squared and cross-products of error terms."
    READ(*,*) cp_option
    WRITE(file_label,*)
!
    IF(cp_option == 0) THEN
       cp_dim = 1
       WRITE(file_label,540)
    ELSE IF(cp_option == 1) THEN
       WRITE(file_label,550)
       cp_dim = p*q*k*k
    ELSE IF(cp_option == 2) THEN
       WRITE(file_label,560)
       cp_dim = (p*k)*(p*k + 1)/2
    ELSE IF(cp_option == 3) THEN
       WRITE(file_label,570)
       cp_dim = (q*k)*(q*k + 1)/2
    END IF
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the coefficients for the second order terms (row by row)."
    WRITE(file_label,580)
    DO dim1 = 1, k
       READ(*,*) (cp_coefs(dim1,dim2), dim2 = 1, cp_dim)
       WRITE(file_label,*) (cp_coefs(dim1,dim2), dim2 = 1, cp_dim)
    END DO
!
400 FORMAT("Vector bilinear process of dimension ",I1,".")
500 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
510 FORMAT("The AR coefficients (row by row) are:")
520 FORMAT("Lag ",I2,":")
530 FORMAT("The MA coefficients (row by row) are:")
540 FORMAT("There are no cross-product terms.")
550 FORMAT("Cross-product terms were the time series with the errors.")
560 FORMAT("Cross-product terms were squared terms and interactions of the time series.")
570 FORMAT("Cross-product terms were squared errors and cross-products of errors.")
580 FORMAT("The second-order term coefficients are (row by row):")
!
    RETURN
  END SUBROUTINE get_mvbilinear_params
!***********************************************************************
!   Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvbilinear(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs,   &
       MA_coefs,cp_option,cp_dim,cp_coefs,cov_fac_option,cov,extra,x)
    USE define_types
    USE f90_utilities, ONLY : vech
    USE stat_utilities
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, max_k, max_p, max_q
    INTEGER(i4b), INTENT(in) :: n, k, p, q, cp_option, cp_dim
    INTEGER(i4b), INTENT(in) :: cov_fac_option, extra
    REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(dp), INTENT(in), DIMENSION(max_k, cp_dim) :: cp_coefs
    REAL(dp), INTENT(in), DIMENSION(max_p, max_k, k) :: AR_coefs
    REAL(dp), INTENT(in), DIMENSION(max_q, max_k, k) :: MA_coefs
    REAL(dp), INTENT(inout), DIMENSION(max_n, k) :: x
!
    REAL(dp), DIMENSION(max_n,k) :: eps
    INTEGER(i4b) :: nn, dim, t, i, j, index1, index2, vech_option, jj, l, dim1, dim2
    REAL(dp), DIMENSION(k) :: work_space
    REAL(dp), DIMENSION((p*k)) :: zv
    REAL(dp), DIMENSION(cp_dim) :: z
    REAL(dp), DIMENSION(1,(p*k)) :: zvt
!
    work_space = 0.0_dp
    eps = 0.0_dp
    nn = n + extra
    dim = max(p,q)*k
    CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = 1, nn
!
!  Terms for computing cross-products:
!
       DO i = 1, k
          DO j = 1, MIN(t-1, p)
             index1 = t - j
             index2 = p*(i-1) + j
             IF(cp_option == 1) THEN
                zv(index2) = x(index1,i)
                zvt(1,index2) = eps(index1,i)
                vech_option = 1
             ELSE IF(cp_option == 2) THEN
                zv(index2) = x(t-j,i)
                zvt(1,index2) = zv(p*(i-1) + j)
                vech_option = 0
             ELSE IF(cp_option == 3) THEN
                zv(index2) = eps(index1,i)
                zvt(1,index2) = zv(index2)
                vech_option = 0
             END IF
             CALL vech(zv,zvt,dim,dim,1,dim,p*k,vech_option,z)
          END DO
       END DO
!
!  Linear AR part:
!
       DO j = 1, MIN(t-1, p)
          DO dim1 = 1, k
          DO dim2 = 1, k
             x(t,dim1) = x(t,dim1) + AR_coefs(j,dim1,dim2)*x(t-j,dim2)
          END DO
          END DO
       END DO
!
!  Linear MA part:
!
       DO j = 1, MIN(t-1, q)
          DO dim1 = 1, k
          DO dim2 = 1, k
             x(t,dim1) = x(t,dim1) - MA_coefs(j,dim1,dim2)*eps(t-j,dim2)
          END DO
          END DO
       END DO
!
!  Nonlinear part:
!
       DO i = 1, k
       DO j = 1, cp_dim
          x(t,i) = x(t,i) + cp_coefs(i,j)*z(j)
       END DO
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):(extra+n),1:k)
!
    RETURN
  END SUBROUTINE gen_mvbilinear
!***********************************************************************
!  Generate vector threshold ARMA:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvTAR_params(max_k,max_p,max_q,max_no_regimes,       &
       file_label,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,         &
       AR_lags,MA_lags,thr_resp_ind,thr_lag,thresholds,TAR_covs)
!***********************************************************************
!  FORTRAN 90 subroutine to get parameters for generating realizations
!  from a multivariate threshold AR process.
!***********************************************************************
    USE define_types
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_k, max_p, max_q, max_no_regimes 
    INTEGER(i4b), INTENT(in) :: file_label, k
    INTEGER(i4b), INTENT(out) :: p, q, no_regimes, thr_resp_ind, thr_lag
    INTEGER(i4b), INTENT(out), DIMENSION(max_p) :: AR_lags
    INTEGER(i4b), INTENT(out), DIMENSION(max_q) :: MA_lags
    REAL(dp), INTENT(out), DIMENSION(max_no_regimes) :: thresholds
    REAL(dp), INTENT(out), DIMENSION(max_no_regimes, max_k, max_k) :: TAR_covs
    REAL(dp), INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: TAR_AR_coefs
    REAL(dp), INTENT(out), DIMENSION(max_no_regimes, max_q, max_k, max_k) :: TAR_MA_coefs
!
    INTEGER(i4b) :: reg_ct, dim1, dim2, info, j
    REAL(dp), DIMENSION(k, k) :: cov
!
    WRITE(file_label,400) k
    WRITE(*,*) "How many regimes in the TAR model?"
    READ(*,*) no_regimes
    WRITE(file_label,500) no_regimes
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,510) p, q
    WRITE(file_label,*)
!
    DO reg_ct = 1, no_regimes
       WRITE(*,520) reg_ct
       WRITE(file_label,520) reg_ct
       DO dim1 = 1,k
          READ(*,*) (cov(dim1,dim2), dim2 = 1, k)
          WRITE(file_label,*) (cov(dim1,dim2), dim2 = 1, k)
       END DO
       TAR_covs(reg_ct,1:k,1:k) = cov(1:k,1:k)
       WRITE(file_label,*)
    END DO
!
    IF(p > 0) THEN
       WRITE(*,*) "What are the AR lags?"
       READ(*,*) (AR_lags(j), j = 1, p)
       WRITE(file_label,525,ADVANCE="NO") 
       WRITE(file_label,*) (AR_lags(j), j = 1, p)
       DO reg_ct = 1, no_regimes
          WRITE(*,530) reg_ct
          WRITE(file_label,540) reg_ct
          DO j = 1, p
             WRITE(*,550) j
             DO dim1 = 1,k
                READ(*,*) (TAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (TAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "What are the MA lags?"
       READ(*,*) (MA_lags(j), j = 1, q)
       WRITE(file_label,*)
       WRITE(file_label,555,ADVANCE="NO") 
       WRITE(file_label,*) (MA_lags(j), j = 1, q)
       DO reg_ct = 1, no_regimes
          WRITE(*,560) reg_ct
          WRITE(file_label,570) reg_ct
          DO j = 1, q
             WRITE(*,580) j
             DO dim1 = 1, k
                READ(*,*) (TAR_MA_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (TAR_MA_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    WRITE(*,590) 
    WRITE(*,*) "Threshold response index?"
    READ(*,*) thr_resp_ind
    WRITE(*,*) "Threshold lag?"
    READ(*,*) thr_lag
    WRITE(file_label,*)
    WRITE(file_label,600) thr_lag, thr_resp_ind
    !
    thresholds(1) = -9.0e15
    thresholds(no_regimes+1) = 9.0e10
    IF(no_regimes > 1) THEN
       WRITE(*,610) no_regimes-1
       READ(*,*) (thresholds(reg_ct+1), reg_ct = 1, no_regimes-1)
       WRITE(file_label,620,ADVANCE="NO")
       WRITE(file_label,*) (thresholds(reg_ct+1), reg_ct = 1, no_regimes-1)
    END IF
!
400 FORMAT("Vector threshold autoregressive moving average model of dimension ",I1,".")
500 FORMAT("The number of regimes in the TAR model is ",I2,".")
510 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
520 FORMAT("Covariance matrix of errors (row by row) for regime ",I2,".")
525 FORMAT("AR lags:")
530 FORMAT("Enter the AR matrix of coefficients for regime ",I2,".")
540 FORMAT("The AR matrix of coefficients for regime ",I2,":")
550 FORMAT("Enter the lag ",I2," AR matrix of coefficients, row by row.")
555 FORMAT("MA lags:")
560 FORMAT("Enter the MA matrix of coefficients for regime ",I2,".")
570 FORMAT("The MA matrix of coefficients for regime ",I2,":")
580 FORMAT("Enter the lag ",I2," MA matrix of coefficients, row by row.")
590 FORMAT("Enter the threshold response index and lag in that order.")
600 FORMAT("The threshold variable is X(t -",I2,",",I2,").")
610 FORMAT("Input ",I2," thesholds.")
620 FORMAT("The thresholds are: ")
!
    RETURN
  END SUBROUTINE get_mvTAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvTAR(max_n,max_k,max_p,max_q,max_no_regimes,          &
       n,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,MA_lags,    &
       thr_resp_ind,thr_lag,thresholds,TAR_covs,cov_fac_option,extra,x)
    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, max_k, max_p, max_q, max_no_regimes
    INTEGER(i4b), INTENT(in) :: n, k, p, q, cov_fac_option, extra
    INTEGER(i4b), INTENT(in) :: no_regimes, thr_resp_ind, thr_lag
    INTEGER(i4b), INTENT(in), DIMENSION(p) :: AR_lags
    INTEGER(i4b), INTENT(in), DIMENSION(q) :: MA_lags
    REAL(dp), INTENT(in), DIMENSION(max_no_regimes) :: thresholds
    REAL(dp), INTENT(in), DIMENSION(max_no_regimes, max_p, max_k, k) :: TAR_AR_coefs
    REAL(dp), INTENT(in), DIMENSION(max_no_regimes, max_q, max_k, k) :: TAR_MA_coefs
    REAL(dp), INTENT(in), DIMENSION(max_no_regimes, max_k, k) :: TAR_covs
    REAL(dp), INTENT(out), DIMENSION (max_n, k) :: x
!
    INTEGER(i4b) :: regime_no, regime, t, j, nn
    REAL(dp) :: crit, thr1, thr2
    REAL(dp), DIMENSION(max_n, k) :: eps
    REAL(dp), DIMENSION(k) :: work_space
    REAL(dp), DIMENSION(max_k, k) :: cov
    REAL(dp), DIMENSION(no_regimes+2,2*n,k) :: all_eps
!
    nn = n + extra
    work_space = 0.0_dp
    DO regime_no = 1, no_regimes
       cov(1:k,1:k) = TAR_covs(regime_no,1:k,1:k)
       CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
       all_eps(regime_no,1:nn,1:k) = eps(1:nn,1:k)
    END DO
!
    work_space = 0.0_dp
    x = 0.0_dp
    DO t = thr_lag+1, nn
       crit = x(t-thr_lag, thr_resp_ind)
       regime = 1
       DO regime_no = 1, no_regimes
          thr1 = thresholds(regime_no)
          thr2 = thresholds(regime_no+1)
          IF(crit .GE. thr1 .AND. crit .LT. thr2) THEN
             regime = regime_no 
             GO TO 10
          END IF
       END DO
10     CONTINUE
!
       x(t,1:k) = all_eps(regime,t,1:k)
       DO j = 1, MIN(t-1, p)
          work_space = MATMUL(x(t-AR_lags(j),1:k),TAR_AR_coefs(regime,j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space
       END DO
       DO j = 1, MIN(t-1, q)
          work_space = MATMUL(all_eps(regime,t-MA_lags(j),1:k),TAR_MA_coefs(regime,j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):nn,1:k)
    RETURN
!
  END SUBROUTINE gen_mvTAR
!***********************************************************************
!  Generate vector threshold STAR:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvSTAR_params(max_k,max_p,max_no_regimes,file_label, &
       k,p,sm_FUNCTION,sm_const,sm_param,STAR_AR_coefs,AR_lags,       &
       thr_resp_ind,thr_lag)
!***********************************************************************
!  FORTRAN 90 subroutine to get parameters for generating realizations
!  from a multivariate smoothed threshold ARMA process.
!***********************************************************************
    USE define_types
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_k, max_p, max_no_regimes
    INTEGER(i4b), INTENT(in) :: file_label, k
    INTEGER(i4b), INTENT(out) :: p, sm_FUNCTION, thr_resp_ind, thr_lag
    INTEGER(i4b), INTENT(out), DIMENSION(max_p) :: AR_lags
    REAL(dp), INTENT(out) :: sm_param, sm_const
    REAL(dp), INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: STAR_AR_coefs
!
    INTEGER(i4b) :: reg_ct, dim1, dim2, info, j
    REAL(dp), DIMENSION(k,k) :: cov
!
    WRITE(file_label,300) k
!
    WRITE(*,*) "What function should be used for smoothing?"
    WRITE(*,*) "  (1 = logistic, 2 = exponential, 3 = quadratic log)"
    READ(*,*) sm_FUNCTION
    WRITE(*,*) "What is the value for the smoothing function constant term?"
    READ(*,*) sm_const
    WRITE(*,*) "What is the value for the smoothing function parameter?"
    READ(*,*) sm_param
    IF(sm_FUNCTION == 1) THEN
       WRITE(file_label,400)
    ELSE IF(sm_FUNCTION == 2) THEN
       WRITE(file_label,410)
    ELSE IF(sm_FUNCTION == 3) THEN
       WRITE(file_label,420)
    END IF
    WRITE(file_label,430) sm_const
    WRITE(file_label,440) sm_param
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the AR order."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(file_label,510) p
    WRITE(file_label,*)
!
    IF(p > 0) THEN
       WRITE(*,*) "What are the AR lags?"
       READ(*,*) (AR_lags(j), j = 1, p)
       WRITE(file_label,525,ADVANCE="NO") 
       WRITE(file_label,*) (AR_lags(j), j = 1, p)
       DO reg_ct = 1, 2
          WRITE(*,530) reg_ct
          WRITE(file_label,540) reg_ct
          DO j = 1, p
             WRITE(*,550) j
             DO dim1 = 1,k
                READ(*,*) (STAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (STAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    WRITE(*,590) 
    WRITE(*,*) "Threshold response index?"
    READ(*,*) thr_resp_ind
    WRITE(*,*) "Threshold lag?"
    READ(*,*) thr_lag
    WRITE(file_label,*)
    WRITE(file_label,600) thr_lag, thr_resp_ind
!
300 FORMAT("Smoothed threshold autoregressive process of dimension ",I1,".")
!
400 FORMAT("The function used for smoothing was the logistic function.")
410 FORMAT("The function used for smoothing was the exponential function.")
420 FORMAT("The function used for smoothing was the quadratic log function.")
430 FORMAT("The smoothing function constant term is ",F8.4,".")
440 FORMAT("The smoothing function parameter is",F8.4,".")
!
510 FORMAT("AR order (p) = ",I2,".")
525 FORMAT("AR lags:")
530 FORMAT("Enter the AR matrix of coefficients for regime ",I2,".")
540 FORMAT("The AR matrix of coefficients for regime ",I2,":")
550 FORMAT("Enter the lag ",I2," AR matrix of coefficients, row by row.")
590 FORMAT("Enter the threshold response index and lag in that order.")
600 FORMAT("The threshold variable is X(t -",I2,",",I2,").")
!
    RETURN
  END SUBROUTINE get_mvSTAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvSTAR(max_n,max_k,max_p,max_no_regimes,n,k,p,         &
       sm_function,sm_const,sm_param,STAR_AR_coefs,AR_lags,             &
       thr_resp_ind,thr_lag,cov,cov_fac_option,extra,x)
    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, max_k, max_p, max_no_regimes
    INTEGER(i4b), INTENT(in) :: n, k, p, cov_fac_option, extra
    INTEGER(i4b), INTENT(in) :: sm_FUNCTION, thr_resp_ind, thr_lag
    INTEGER(i4b), INTENT(in), DIMENSION(p) :: AR_lags
    REAL(dp), INTENT(in) :: sm_const, sm_param
    REAL(dp), INTENT(in), DIMENSION(max_no_regimes, max_p, max_k, k) :: STAR_AR_coefs
    REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(dp), INTENT(out), DIMENSION (max_n, k) :: x
!
    INTEGER(i4b) :: nn, regime_no, regime, t, j
    REAL(dp) :: s
    REAL(dp), DIMENSION(max_n, k) :: eps
    REAL(dp), DIMENSION(k) :: work_space1, work_space2
!
    nn = n + extra
    work_space1 = 0.0_dp
    eps = 0.0_dp
    CALL gen_normal(max_n,max_k,nn,k,work_space1,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = thr_lag+1, nn
       work_space1 = 0.0_dp
       work_space2 = 0.0_dp
!
       DO j = 1, MIN(t-1, p)
          work_space1 = MATMUL(x(t-AR_lags(j),1:k),STAR_AR_coefs(1,j,1:k,1:k))
       END DO
       DO j = 1, MIN(t-1, p)
          work_space2 = MATMUL(x(t-AR_lags(j),1:k),STAR_AR_coefs(2,j,1:k,1:k))
       END DO
!
       s = 0.0_dp
       IF(sm_FUNCTION == 1) THEN
          s = logistic_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       ELSE IF(sm_FUNCTION == 2) THEN
          s = exponential_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       ELSE IF(sm_FUNCTION == 3) THEN
          s = quad_log_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       END IF
!
       x(t,1:k) = work_space1*(1.0_dp - s) + work_space2*s + eps(t,1:k)
    END DO
!
    x(1:n,1:k) = x((extra+1):nn,1:k)
!
    RETURN
  END SUBROUTINE gen_mvSTAR
!***********************************************************************
!  Generate vector exponential AR:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvEXPAR_params(max_k,max_p,file_label,k,p,AR_cons,      &
       AR_coefs,expar_coefs,d)
!***********************************************************************
!  FORTRAN90 subroutine to get parameters for multivariate exponential
!  autoregressive process.
!***********************************************************************
    USE define_types
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_k, max_p, file_label, k
    INTEGER(i4b), INTENT(out) :: p, d
    REAL(dp), INTENT(out), DIMENSION(max_p, max_k, max_k) :: AR_cons, AR_coefs
    REAL(dp), INTENT(out), DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
!
    INTEGER(i4b) :: ts_lag, dim1, dim2, i
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p)."
    READ(*,*) p
    WRITE(file_label,500) p
    WRITE(file_label,*)
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR intercept terms (row by row)."
       WRITE(file_label,510)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             READ(*,*) (AR_cons(ts_lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_cons(ts_lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
!
       WRITE(*,*) "Enter the AR coefficients (row by row)."
       WRITE(file_label,*)
       WRITE(file_label,530)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(ts_lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(ts_lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
!
       WRITE(*,*) "Enter the delay for the exponential term."
       READ(*,*) d
       WRITE(file_label,550) d

       WRITE(*,*) "Enter the exponential parameters (row by row)."
       WRITE(file_label,*)
       WRITE(file_label,540)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             WRITE(*,525) dim1
             WRITE(file_label,525) dim1
             READ(*,*) (expar_coefs(ts_lag,dim2,dim1), dim2 = 1, k*k)
             WRITE(file_label,*) (expar_coefs(ts_lag,dim2,dim1), dim2 = 1, k*k)
          END DO
       END DO
!
    END IF
!
400 FORMAT("Vector exponential autoregressive process of dimension ",I1,".")
500 FORMAT("The AR order (p) is ",I2,".")
510 FORMAT("The AR constant terms (row by row):")
520 FORMAT("Lag ",I2,":")
525 FORMAT("Time series component ",I1,":")
530 FORMAT("The AR coefficients (row by row) are:")
540 FORMAT("The exponential parameters (row by row) are:")
550 FORMAT("Delay for the exponential variable: ",I1,".")
!
    RETURN
  END SUBROUTINE get_mvEXPAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvEXPAR(max_n,max_k,max_p,n,k,p,AR_cons,AR_coefs,      &
       expar_coefs,d,cov,cov_fac_option,extra,x)
    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, max_k, max_p, n, k, p, d
    INTEGER(i4b), INTENT(in) :: cov_fac_option, extra
    REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(dp), INTENT(in), DIMENSION(max_p, max_k, k) :: AR_cons, AR_coefs
    REAL(dp), INTENT(in), DIMENSION(max_p, max_k*max_k, k) :: expar_coefs
    REAL(dp), INTENT(out), DIMENSION(max_n, k) :: x
!
    INTEGER(i4b) :: nn, s, t, j, dim1, dim2, i, ii
    REAL(dp), DIMENSION(max_n, k) :: eps
    REAL(dp), DIMENSION(k) :: work_space, temp
    REAL(dp), DIMENSION(k, k) :: cc
    REAL(dp), DIMENSION(p*k, k) :: eterm
    REAL(dp), DIMENSION(p, k, k) :: coefs
!
    nn = n + extra
    work_space = 0.0_dp
    eterm = 0.0_dp
    temp = 0.0_dp
    eps = 0.0_dp
    s = max(p,d)
    CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = s+1, nn
       coefs = 0.0_dp
       temp(1:k) = x(t-d,1:k)*x(t-d,1:k)
       DO j = 1, p	   
          DO dim1 = 1, k	   
             DO dim2 = 1, k
                i = k*(dim2-1) + 1
                ii = dim2*k
                work_space(1:k) = expar_coefs(j,i:ii,dim1)
                eterm(dim1,dim2) = DOT_PRODUCT(temp,work_space)
             END DO
          END DO
          coefs(j,1:k,1:k) = AR_coefs(j,1:k,1:k)*EXP(eterm(1:k,1:k))
          coefs(j,1:k,1:k) = coefs(j,1:k,1:k) + AR_cons(j,1:k,1:k)
!
          cc(1:k,1:k) = coefs(j,1:k,1:k)
          cc = TRANSPOSE(cc)
!
          work_space(1:k) = MATMUL(x(t-j,1:k),cc)
          x(t,1:k) = x(t,1:k) + work_space(1:k)
       END DO
    END DO
!  
    x(1:n,1:k) = x((extra+1):nn,1:k)
!
    RETURN
  END SUBROUTINE gen_mvEXPAR
!
END MODULE nlts_utilities
